# Adapted from Genevieve Hayes blog
# https://towardsdatascience.com/getting-started-with-reinforcement-learning-and-open-ai-gym-c289aca874f
#

import numpy as np
import gym
import matplotlib.pyplot as plt

plt.style.use('fivethirtyeight')

env = gym.make('MountainCar-v0')
env.reset()

def QLearning(env, learning, discount, epsilon, min_eps, episodes, viz=True):
    # Determine size of discretized state space
    num_states = (env.observation_space.high - env.observation_space.low)*\
                    np.array([10, 100])
    num_states = np.round(num_states, 0).astype(int) + 1

    # Initialize Q table
    Q = np.random.uniform(low = -1, high = 1, size = (num_states[0], num_states[1], env.action_space.n))

    # Initialize variables to track rewards
    reward_list = []
    ave_reward_list = []
    raw_rewards = []

    # Calculate episodic reduction in epsilon
    reduction = (epsilon - min_eps) / episodes

    for i in range(episodes):
        episode_terminated = False
        tot_reward, reward = 0, 0
        state = env.reset()

        # Discretize state
        state_adj = (state - env.observation_space.low)*np.array([10, 100])
        state_adj = np.round(state_adj, 0).astype(int)

        while not episode_terminated:
            # Render environment for last five episodes
            if viz and i >= (episodes - 5):
                env.render()

            # Determine next action - epsilon greedy strategy
            if np.random.random() < 1 - epsilon:
                action = np.argmax(Q[state_adj[0], state_adj[1]])
            else:
                action = np.random.randint(0, env.action_space.n)

            # Get next state and reward
            state2, reward, episode_terminated, info = env.step(action)

            # Discretize state2
            state2_adj = (state2 - env.observation_space.low)*np.array([10, 100])
            state2_adj = np.round(state2_adj, 0).astype(int)

            #Allow for terminal states
            if episode_terminated and state2[0] >= 0.5:
                Q[state_adj[0], state_adj[1], action] = reward
            else:
                delta = learning*(reward + discount*np.max(Q[state2_adj[0], state2_adj[1]]) - Q[state_adj[0], state_adj[1],action])
                Q[state_adj[0], state_adj[1],action] += delta

            # Update variables
            tot_reward += reward
            state_adj = state2_adj

        # Decay epsilon
        if epsilon > min_eps:
            epsilon -= reduction

        # Track rewards
        reward_list.append(tot_reward)

        if (i+1) % 100 == 0:
            ave_reward = np.mean(reward_list)
            ave_reward_list.append(ave_reward)
            raw_rewards.append(reward_list)
            reward_list = []

        if (i+1) % 100 == 0:
            print('Episode {} Average Reward: {}'.format(i+1, ave_reward))

    env.close()

    return ave_reward_list, np.array(raw_rewards)

# Run Q-learning algorithm
avr_rewards, raw_rewards = QLearning(env, 0.25, 0.95, 0.8, 0, 5000)
# Plot Rewards
f = plt.figure(figsize=(14, 8))
plt.plot(100*(np.arange(len(avr_rewards)) + 1), avr_rewards)
plt.xlabel('Episodes')
plt.ylabel('Average Reward')
plt.title('Average Reward vs Episodes')
plt.savefig('avr_rewards.pdf')

f = plt.figure(figsize=(14, 8))
plt.plot(raw_rewards, alpha=0.5, ls='', marker='o')
plt.xlabel('Time step')
plt.ylabel('Reward')
plt.title('Episodic Rewards')
plt.savefig('episode_rewards.pdf')

plt.show()


# # Multiple.
# f = plt.figure(figsize=(14, 8))
# for i in range(20):
#     a_r, _ = QLearning(env, 0.25, 0.95, 0.8, 0, 10000, viz=False)
#     plt.plot(100*(np.arange(len(a_r)) + 1), a_r, alpha=0.7)
#     plt.xlabel('Episodes')
#     plt.ylabel('Average Reward')
#     plt.title('Average Reward vs Episodes [Multiple runs]')
# plt.savefig('avr_rewards_multi.pdf')
#
# plt.show()
